from typing import Type, Tuple

from wai.annotations.core.component import ProcessorComponent
from wai.annotations.core.domain import DomainSpecifier
from wai.annotations.core.specifier import ProcessorStageSpecifier


class PredictISPSpecifier(ProcessorStageSpecifier):
    """
    Specifies the object detection predict ISP.
    """
    @classmethod
    def description(cls) -> str:
        return "Makes object detection predictions via Redis backend, passing in an image and receiving OPEX predictions back:\n" \
               + "https://github.com/WaikatoLink2020/objdet-predictions-exchange-format"

    @classmethod
    def domain_transfer_function(
            cls,
            input_domain: Type[DomainSpecifier]
    ) -> Type[DomainSpecifier]:
        from wai.annotations.domain.image.object_detection import ImageObjectDetectionDomainSpecifier
        if input_domain is ImageObjectDetectionDomainSpecifier:
            return input_domain
        else:
            raise Exception(
                f"Predict only handles the following domains: "
                f"{ImageObjectDetectionDomainSpecifier.name()}"
            )

    @classmethod
    def components(cls) -> Tuple[Type[ProcessorComponent]]:
        from wai.annotations.redis_predictions.isp.object_detection.component import Predict
        return Predict,



from wai.annotations.core.component import ProcessorComponent
from wai.annotations.core.stream import ThenFunction, DoneFunction
from wai.annotations.domain.image.object_detection import ImageObjectDetectionInstance
from wai.annotations.redis_predictions.isp.base.component import BasePredict
from wai.common.adams.imaging.locateobjects import LocatedObjects, LocatedObject
from wai.common.cli.options import TypedOption
from wai.common.geometry import Polygon, Point
from opex import ObjectPredictions


class Predict(
    BasePredict,
    ProcessorComponent[ImageObjectDetectionInstance, ImageObjectDetectionInstance]
):
    """
    Stream processor which makes object detection predictions via Redis backend.
    """

    key_label: str = TypedOption(
        "--key-label",
        type=str,
        default="type",
        help="the meta-data key in the annotations to use for storing the label."
    )

    key_score: str = TypedOption(
        "--key-score",
        type=str,
        default="score",
        help="the meta-data key in the annotations to use for storing the prediction score."
    )

    def _process_predictions(self, element, data, then: ThenFunction[ImageObjectDetectionInstance], done: DoneFunction):
        """
        Processes the prediction data.

        :param element: the incoming image
        :param data: the data to process
        :param then: the function to call with the parsed prediction data
        :param done: if necessary to call
        :return:
        """
        # convert to wai.annotations annotations
        oobjects = ObjectPredictions.from_json_string(data)
        lobjects = []
        for oobject in oobjects.objects:
            # bbox
            obbox = oobject.bbox
            x = obbox.left
            y = obbox.top
            w = obbox.right - obbox.left + 1
            h = obbox.bottom - obbox.top + 1

            # polygon
            opoly = oobject.polygon
            lpoints = []
            for opoint in opoly.points:
                lpoints.append(Point(x=opoint[0], y=opoint[1]))
            lpoly = Polygon(*lpoints)

            # metadata
            metadata = dict()
            if hasattr(oobject, "score"):
                metadata[self.key_score] = oobject.score
            metadata[self.key_label] = oobject.label

            # add object
            located_object = LocatedObject(x, y, w, h, **metadata)
            located_object.set_polygon(lpoly)
            lobjects.append(located_object)

        annotations = LocatedObjects(lobjects)

        # update image
        result = element.__class__(element.data, annotations)
        then(result)



from typing import Type, Tuple

from wai.annotations.core.component import ProcessorComponent
from wai.annotations.core.domain import DomainSpecifier
from wai.annotations.core.specifier import ProcessorStageSpecifier


class PredictISPSpecifier(ProcessorStageSpecifier):
    """
    Specifies the annotation_overlay image ISP.
    """
    @classmethod
    def description(cls) -> str:
        return "Makes image classification predictions via Redis backend, passing in an image and receiving JSON predictions back (at least one of 'label: probability').\n" \
               + "Predictions example:\n" \
               + "{\"dog\": 0.9, \"cat\": 0.1}"

    @classmethod
    def domain_transfer_function(
            cls,
            input_domain: Type[DomainSpecifier]
    ) -> Type[DomainSpecifier]:
        from wai.annotations.domain.image.classification import ImageClassificationDomainSpecifier
        if input_domain is ImageClassificationDomainSpecifier:
            return input_domain
        else:
            raise Exception(
                f"Predict only handles the following domains: "
                f"{ImageClassificationDomainSpecifier.name()}"
            )

    @classmethod
    def components(cls) -> Tuple[Type[ProcessorComponent]]:
        from wai.annotations.redis_predictions.isp.image_classification.component import Predict
        return Predict,



import redis

from datetime import datetime
from time import sleep

from wai.common.cli.options import TypedOption, FlagOption
from wai.annotations.core.component import ProcessorComponent
from wai.annotations.core.stream import ThenFunction, DoneFunction
from wai.annotations.core.stream.util import RequiresNoFinalisation
from wai.annotations.domain.image import ImageInstance


class DataContainer(object):
    """
    Simple data container.
    """

    def __init__(self):
        self.redis: redis.Redis = None
        self.channel_out: str = "images"
        self.channel_in: str = "predictions"
        self.timeout: float = 5.0
        self.data = None


class BasePredict(
    RequiresNoFinalisation,
    ProcessorComponent[ImageInstance, ImageInstance]
):
    """
    Base class for Stream processors that make predictions via Redis backend.
    """

    redis_host: str = TypedOption(
        "-h", "--redis-host",
        type=str,
        default="localhost",
        help="the Redis server to connect to"
    )

    redis_port: int = TypedOption(
        "-p", "--redis-port",
        type=int,
        default=6379,
        help="the port the Redis server is running on"
    )

    redis_db: int = TypedOption(
        "-d", "--redis-db",
        type=int,
        default=0,
        help="the database to use"
    )

    channel_out: str = TypedOption(
        "--channel-out",
        type=str,
        default="images",
        help="the Redis channel to send the images out"
    )

    channel_in: str = TypedOption(
        "--channel-in",
        type=str,
        default="predictions",
        help="the Redis channel on which to receive predictions."
    )

    timeout: float = TypedOption(
        "-t", "--timeout",
        type=float,
        default=5.0,
        help="the timeout in seconds to wait for a prediction to arrive"
    )

    verbose: bool = FlagOption(
        "-v", "--verbose",
        help="whether to output debugging information."
    )

    def _process_predictions(self, element, data, then, done):
        """
        Processes the prediction data.

        :param element: the incoming image
        :param data: the data to process
        :param then: the function to call with the parsed prediction data
        :param done: if necessary to call
        :return:
        """
        raise NotImplementedError()

    def process_element(
            self,
            element: ImageInstance,
            then: ThenFunction[ImageInstance],
            done: DoneFunction
    ):
        if not hasattr(self, "_redis_conn"):
            self._redis_conn = DataContainer()
            self._redis_conn.redis = redis.Redis(host=self.redis_host, port=self.redis_port, db=self.redis_db)
            self._redis_conn.channel_out = self.channel_out
            self._redis_conn.channel_in = self.channel_in
            self._redis_conn.timeout = self.timeout
            self._redis_conn.data = None

        def anon_handler(message):
            data = message['data']
            self._redis_conn.data = data
            self._redis_conn.pubsub_thread.stop()
            self._redis_conn.pubsub.close()
            self._redis_conn.pubsub = None

        self._redis_conn.pubsub = self._redis_conn.redis.pubsub()
        self._redis_conn.pubsub.psubscribe(**{self._redis_conn.channel_in: anon_handler})
        self._redis_conn.pubsub_thread = self._redis_conn.pubsub.run_in_thread(sleep_time=0.001)
        self._redis_conn.redis.publish(self._redis_conn.channel_out, element.data.data)

        # wait for data to show up
        start = datetime.now()
        no_data = False
        while self._redis_conn.pubsub is not None:
            sleep(0.001)
            end = datetime.now()
            if self._redis_conn.timeout > 0:
                if (end - start).total_seconds() >= self._redis_conn.timeout:
                    if self.verbose:
                        self.logger.info("Timeout reached!")
                    no_data = True
                    break

        if no_data:
            return
        elif self.verbose:
            end = datetime.now()
            self.logger.info("Round trip time: %f sec" % (end - start).total_seconds())

        # process predictions
        self._process_predictions(element, self._redis_conn.data, then, done)



import io
import numpy as np

from PIL import Image

from wai.annotations.core.component import ProcessorComponent
from wai.annotations.core.stream import ThenFunction, DoneFunction
from wai.annotations.domain.image.segmentation import ImageSegmentationInstance, ImageSegmentationAnnotation
from wai.annotations.domain.image.segmentation.util import UnlabelledInputMixin
from wai.annotations.redis_predictions.isp.base.component import BasePredict
from wai.common.cli.options import TypedOption

FORMAT_INDEXEDPNG = "indexedpng"
FORMAT_BLUECHANNEL = "bluechannel"
FORMATS = [
    FORMAT_INDEXEDPNG,
    FORMAT_BLUECHANNEL,
]


class Predict(
    BasePredict,
    UnlabelledInputMixin,
    ProcessorComponent[ImageSegmentationInstance, ImageSegmentationInstance]
):
    """
    Stream processor which makes image segmentation predictions via Redis backend.
    """

    image_format: str = TypedOption(
        "--image-format",
        type=str,
        default=FORMAT_INDEXEDPNG,
        help="the format of the image that comes back as prediction: %s" % ",".join(FORMATS)
    )

    def _fix_size(self, img, width, height):
        """
        Fixes the size of the received image, if necessary.

        :param img: the to resize
        :type img: Image
        :param width: the required width
        :type width: int
        :param height: the required height
        :type height: int
        :return: the (potentially) resized image
        :rtype: Image
        """
        if (img.width == width) and (img.height == height):
            return img
        else:
            return img.resize(width, height, Image.BILINEAR)

    def _process_predictions(self, element, data, then: ThenFunction[ImageSegmentationInstance], done: DoneFunction):
        """
        Processes the prediction data.

        :param element: the incoming image
        :param data: the data to process
        :param then: the function to call with the parsed prediction data
        :param done: if necessary to call
        :return:
        """

        annotations = ImageSegmentationAnnotation(labels=self.labels, size=element.annotations.size)

        w = element.data.width
        h = element.data.height

        # convert received image to indices
        if self.image_format == FORMAT_INDEXEDPNG:
            image = self._fix_size(Image.open(io.BytesIO(data)), w, h)
            new_indices = np.asarray(image).astype(np.uint16)
            new_indices.resize(element.annotations.indices.shape)
            annotations.indices = new_indices
        elif self.image_format == FORMAT_BLUECHANNEL:
            image = self._fix_size(Image.open(io.BytesIO(data)), w, h)
            new_indices = np.asarray(image).astype(np.uint16)
            new_indices.resize((*element.annotations.indices.shape, 3))
            new_indices = new_indices[:, :, 2]
            new_indices = new_indices.astype(np.uint16)
            annotations.indices = new_indices
        else:
            raise Exception("Unsupported image format: %s" % self.image_format)

        # update image
        result = element.__class__(element.data, annotations)
        then(result)



import json

from wai.annotations.core.component import ProcessorComponent
from wai.annotations.core.stream import ThenFunction, DoneFunction
from wai.annotations.domain.image.classification import ImageClassificationInstance
from wai.annotations.domain.classification import Classification
from wai.annotations.redis_predictions.isp.base.component import BasePredict


class Predict(
    BasePredict,
    ProcessorComponent[ImageClassificationInstance, ImageClassificationInstance]
):
    """
    Stream processor which makes image classification predictions via Redis backend.
    """

    def _process_predictions(self, element, data, then: ThenFunction[ImageClassificationInstance], done: DoneFunction):
        """
        Processes the prediction data.

        :param element: the incoming image
        :param data: the data to process
        :param then: the function to call with the parsed prediction data
        :param done: if necessary to call
        :return:
        """
        # convert to wai.annotations annotations
        preds = json.loads(data)
        max_key = None
        max_value = 0.0
        for k in preds:
            if preds[k] > max_value:
                max_key = k
                max_value = preds[k]

        annotations = Classification(label=max_key)

        # update image
        result = element.__class__(element.data, annotations)
        then(result)
